/*
 *  aud_scan.c
 *
 *  Scans the audio track
 *
 *  Copyright (C) Tilmann Bitterberg - June 2003
 *
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>

//#include "avilib.h"

#define EXE	"ac3info"
#define VERSION "0.1"

typedef struct {
	int nchans;
	char *chan_ordering;
	int lfe;
	int bitrate;
	int srate;
	char *filename;
	char *bsmod;
	char *cmixlev;
	char *surmixlev;
	char *surmode;
	int acmod;
} ac3_info;

static char *bsmod_table[] = { "main audio service: complete main     (CM)",
			  "main audio service: music and effects (ME)",
			  "associated service: visually impaired (VI)",
			  "associated service: hearing impaired  (HI)",
			  "associated service: dialogue          (D)",
			  "associated service: commentary	 (C)",
			  "associated service: emergency	 (E)",
			  "associated service: voice over	 (VO)",
			  "main audio service: karaoke" };

static const unsigned char nfchans[] = {2,1,2,3,3,4,4,5,1,1,2};
static char *chan_ordering[] = {"Ch1, Ch2","C","L,R","L,C,R","L,R,S","L,C,R,S","L,R,SL,SR","L,C,R,SL,SR","","",""};
static char *cmix_lev[] = {"-3.0 dB", "-4.5 dB", "-6.0 dB", "reserved"};
static char *surmix_lev[] = {"-3.0 dB", "-6.0 dB", "0.0 dB", "reserved"};
static char *surmode_list[] = {"not indicated", "Not Dolby Surround Encoded", "Dolby Surround Encoded", "reserved"};

struct frmsize_s
{
	uint16_t bit_rate;
	uint16_t frm_size[3];

} frmsize_t;

static const struct frmsize_s frmsizecod_tbl[] = 
{
	{ 32  ,{64   ,69   ,96   } },
	{ 32  ,{64   ,70   ,96   } },
	{ 40  ,{80   ,87   ,120  } },
	{ 40  ,{80   ,88   ,120  } },
	{ 48  ,{96   ,104  ,144  } },
	{ 48  ,{96   ,105  ,144  } },
	{ 56  ,{112  ,121  ,168  } },
	{ 56  ,{112  ,122  ,168  } },
	{ 64  ,{128  ,139  ,192  } },
	{ 64  ,{128  ,140  ,192  } },
	{ 80  ,{160  ,174  ,240  } },
	{ 80  ,{160  ,175  ,240  } },
	{ 96  ,{192  ,208  ,288  } },
	{ 96  ,{192  ,209  ,288  } },
	{ 112 ,{224  ,243  ,336  } },
	{ 112 ,{224  ,244  ,336  } },
	{ 128 ,{256  ,278  ,384  } },
	{ 128 ,{256  ,279  ,384  } },
	{ 160 ,{320  ,348  ,480  } },
	{ 160 ,{320  ,349  ,480  } },
	{ 192 ,{384  ,417  ,576  } },
	{ 192 ,{384  ,418  ,576  } },
	{ 224 ,{448  ,487  ,672  } },
	{ 224 ,{448  ,488  ,672  } },
	{ 256 ,{512  ,557  ,768  } },
	{ 256 ,{512  ,558  ,768  } },
	{ 320 ,{640  ,696  ,960  } },
	{ 320 ,{640  ,697  ,960  } },
	{ 384 ,{768  ,835  ,1152 } },
	{ 384 ,{768  ,836  ,1152 } },
	{ 448 ,{896  ,975  ,1344 } },
	{ 448 ,{896  ,976  ,1344 } },
	{ 512 ,{1024 ,1114 ,1536 } },
	{ 512 ,{1024 ,1115 ,1536 } },
	{ 576 ,{1152 ,1253 ,1728 } },
	{ 576 ,{1152 ,1254 ,1728 } },
	{ 640 ,{1280 ,1393 ,1920 } },
	{ 640 ,{1280 ,1394 ,1920 } }
};

#define fscd_tbl_entries (sizeof(frmsizecod_tbl)/sizeof(frmsize_t))

void version()
{
  printf("%s v%s (C) 2007 Erwan Velu\n", EXE, VERSION);
}

void usage(int status)
{
    version();
    printf("\nUsage: %s [options]\n", EXE);
    printf("\t file                 Ac3 file name\n");
    exit(status);
}

unsigned long get_ac3_header(unsigned char *buf) 
{
  int i=0;
  unsigned long tmp=0;

  tmp = (tmp << 8) + (buf[i++]&0xff);
  tmp = (tmp << 8) + (buf[i++]&0xff);
  tmp = (tmp << 8) + (buf[i++]&0xff);
  
  return(tmp);
}

int get_ac3_framesize(unsigned char *buf) 
{
  int fscod, frmsizecod;
  unsigned long tmp = 0;

  tmp=get_ac3_header(buf);
  
  if(tmp<0) return(-1);
  
  fscod      = (tmp >> 6) & 0x3;
  frmsizecod = tmp & 0x3f;

  if(frmsizecod >= fscd_tbl_entries || fscod > 2) return(-1);

  return(frmsizecod_tbl[frmsizecod].frm_size[fscod]);
}

// We try to find the number of chans in the ac3 header (BSI)
int get_ac3_nfchans(unsigned char *buf, ac3_info *ac3_info) 
{
  int acmod = 0;
  int bsmod = 0;
 
  /* acmod is located on the 3 msb of the second char of the BSI */
  acmod = buf[1]>>5;
 
  /* bsmod is located on the 3 lsb of the first char of the BSI */
  bsmod = buf[0] & 0x07;
  /* According to the 5.4.2.2 of the AC3 doc,
   * if bsmod is 111, and acmod >1 the bsmod is karaoke */
  if ((bsmod == 0x07) && ( acmod > 1)) {
	  bsmod=0x08;
  }
  ac3_info->bsmod=bsmod_table[bsmod];
  
  /* LFE is the 2nd msb bit (0x40) of the 3rd char of the BSI */
  if ((buf[2] & 0x40) == 0x40) {
  	  /* LFE flags is on, we have one more channel */
	  ac3_info->lfe=1;
  }
  
  if (acmod < 0 || acmod > 11) return -1;
  
  ac3_info->acmod=acmod;
  ac3_info->nchans=nfchans[acmod];
  ac3_info->chan_ordering=chan_ordering[acmod];
  return(ac3_info->nchans);
}

int get_ac3_cmixlev(unsigned char *buf, ac3_info *ac3_info) {
  int cmix;
  ac3_info->cmixlev=NULL;
  /* If 3 front channels*/
  if ((ac3_info->acmod & 0x1) && (ac3_info->acmod != 0x1)) {
	  /* cmixlev are the 4th msb and 1st lsb */
	  cmix=(buf[1] & 0x18) >> 3;
	  ac3_info->cmixlev=cmix_lev[cmix];
  }
}

int get_ac3_surmixlev(unsigned char *buf, ac3_info *ac3_info) {
  int surmix;
  ac3_info->surmixlev=NULL;
  /* If a surround channel exist */
  if (ac3_info->acmod & 0x4) {
	  /* surmixlev are at the 2 & 3 third lsb */
	  surmix=(buf[1] & 0x06) >> 1;
	  ac3_info->surmixlev=surmix_lev[surmix];
  }
}

int get_ac3_surmode(unsigned char *buf, ac3_info *ac3_info) {
  int surmode,surmode1,surmode2;
  /* If we are in 2.0 mode */
  ac3_info->surmode=NULL;
  if (ac3_info->acmod == 0x2) {
	  surmode1=(buf[1] & 0x01) << 1;
	  surmode2=(buf[2] & 0x80) >> 7;
	  surmode= surmode1 | surmode2;
	  ac3_info->surmode=surmode_list[surmode];
  }
}
int get_ac3_bitrate(unsigned char *buf) 
{
  int frmsizecod;
  unsigned long tmp = 0;

  tmp=get_ac3_header(buf);
  
  frmsizecod = tmp & 0x3f;

  if(frmsizecod >= fscd_tbl_entries) return(-1);

  return(frmsizecod_tbl[frmsizecod].bit_rate);
}


int get_ac3_samplerate(unsigned char *buf) 
{
  int fscod, sampling_rate;
  unsigned long tmp = 0;

  tmp=get_ac3_header(buf);
  
  // Get the sampling rate 
  fscod  = (tmp >> 6) & 0x3;
  
  if(fscod == 3) {
    return(-1);  //invalid sampling rate code
  } else if(fscod == 2)
    sampling_rate = 32000;
  else if(fscod == 1)
    sampling_rate = 44100;
  else
    sampling_rate = 48000;

  return(sampling_rate);
}


int tc_get_ac3_header(unsigned char *_buf, int len, ac3_info *ac3_info)
{

  int j=0, i=0, fsize, nfchans;

  unsigned char *buffer;

  uint16_t sync_word = 0;

  // need to find syncframe:

  buffer=_buf;
  
  for (i=0; i<len-4; i++) {
      //fprintf(stderr, "%02x ", buffer[i]);
      sync_word = (sync_word << 8) + (uint8_t) buffer[i]; 
      if(sync_word == 0x0b77) break;
  }
  if(sync_word != 0x0b77) return(-1);

  ac3_info->srate = get_ac3_samplerate(&buffer[i+1]);
  ac3_info->bitrate = get_ac3_bitrate(&buffer[i+1]);  

  /* We give the first byte of the BSI to get_ac3_nfchans*/
  get_ac3_nfchans(&buffer[i+4],ac3_info);

  get_ac3_cmixlev(&buffer[i+4],ac3_info);
  get_ac3_surmixlev(&buffer[i+4],ac3_info);
  get_ac3_surmode(&buffer[i+4],ac3_info);
  fsize = 2*get_ac3_framesize(&buffer[i+1]);
  
  if(j<0 || ac3_info->bitrate <0) return(-1);

  return(fsize);
}

void display_ac3_info(ac3_info *ac3) {
  printf("AC3 Informations for %s\n",ac3->filename);
  printf("Basic Informations\n");
  printf("- Channels           : %4d.%d\n",ac3->nchans,ac3->lfe);
  printf("- Sample Rate        : %6d Khz\n",ac3->srate);
  printf("- Bitrate            : %6d Kbits/sec\n",ac3->bitrate);
  printf("\nAdvanced Informations\n");
  printf("- Bit Stream Mode    : %s \n",ac3->bsmod); 
  printf("- Channels Ordering  : %s\n",ac3->chan_ordering);
  if (ac3->cmixlev != NULL) {
  printf("- Cmix level         : %s\n",ac3->cmixlev);
  }
  if (ac3->surmixlev != NULL) {
  printf("- Surround Mix level : %s\n",ac3->surmixlev);
  }
  if (ac3->surmode != NULL) {
  printf("- Surround Mode      : %s\n",ac3->surmode);
  }

}

int parse_ac3(ac3_info *ac3) {
  FILE *f;
  unsigned char buffer[8];
  int len;
  int exit_status=0;
  int result;
  
  f = fopen(ac3->filename,"rb");
  if (!f) { perror ("fopen"); exit(1); }
  
  len=fread(buffer,1,8,f);
  if (len!=8) {
	  printf("%s is not readable or too small\n",ac3->filename);
	  exit_status=-1;
	  goto exit;
  }
  result=tc_get_ac3_header(buffer,len,ac3);
  if (result>=0) {
	display_ac3_info(ac3);
  } else {
	printf("%s is not an AC3 file\n",ac3->filename); 
	goto exit;
  }

exit:
  fclose(f);
  return exit_status;
}

int main(int argc, char *argv[]) {
  int ch;
  ac3_info ac3;
  if(argc==1) usage(EXIT_FAILURE);
  
  ac3.filename=argv[1];
  parse_ac3(&ac3);
}
